#!/usr/bin/perl

package LocIP::DNS;

use strict;
use Net::DNS;
use Net::DNS::RR;
use Socket;
use Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(&localize_by_loc_dns &localize_by_txt_dns &localize_by_hostname);


# funkcia lokalizuje IP adresu pouzitim LOC vety DNS zanamu
# vstupne parametre: $hostname - domenove meno uzla
#                    $dbh      - deskriptor databazoveho pripojenia na databazu so suradnicami 
# navratova hodnota: pole (zoznam) lokalizacnych udajov  ($country, $city, $lat, $lon)
#                    $country - krajina lokalizacie
#                    $city    - mesto lokalizacie
#                    $lat     - latitude (zemepisna sirka)
#                    $lon     - longitude (zemepisna dlzka)
sub localize_by_loc_dns {
    my ($hostname, $dbh) = @_;
    my ($country, $city, $lat, $lon);
    
    if (! $hostname) { return; } # ak nie je zadane domenove meno, ukonci funkciu
    
    # prechadza vsetkymy urovnami domeny:
    my $count = 0;
    while ($hostname =~ /\./) { # ak host obsahuje bodku
        if (($lat, $lon) = get_dns_record($hostname, "LOC")) { # pokusi sa ziskat zemepisnu sirku a dlzku z LOC DNS zaznamu
            last;
        }
        ($hostname) = $hostname =~ /\.(.*)$/; # odtrhne z hostu cast po prvu bodku zlava (vratane bodky)
        if (++$count == 2) { last; }  # obmedzenie poctu DNS requestov na 2
    }
    
    if ($lat) {           
        # pokusi sa zistit mesto a krajinu zo suradnic:
        my $sel = $dbh->prepare("SELECT country, city, sqrt((latitude-?)*(latitude-?)+(longitude-?)*(longitude-?)) AS distance FROM geo_city WHERE ROUND(latitude, 1) = ROUND(?, 1) AND ROUND(longitude, 1) = ROUND(?, 1) ORDER BY distance LIMIT 1");
		$sel->execute($lat, $lat, $lon, $lon, $lat, $lon);
        while (my @row = $sel->fetchrow_array) {
            ($country, $city) = @row; # ulozi krajinu a mesto z databazy do premennych
        }
           
    }
    return ($country, $city, $lat, $lon); 
}


# funkcia lokalizuje IP adresu pouzitim TXT vety DNS zanamu
# vstupne parametre: $hostname        - domenove meno uzla
#                    $dbh             - deskriptor databazoveho pripojenia na databazu so suradnicami
#                    $default_country - krajina, ktora uz bola urcena inymi sposobmi
# navratova hodnota: pole (zoznam) lokalizacnych udajov  ($country, $city, $lat, $lon)
#                    $country - krajina lokalizacie
#                    $city    - mesto lokalizacie
#                    $lat     - latitude (zemepisna sirka)
#                    $lon     - longitude (zemepisna dlzka)
sub localize_by_txt_dns {
    my ($hostname, $dbh, $default_country) = @_;
    my ($country, $city, $lat, $lon, $txt);
    
    if (! $hostname) { return; } # ak nie je zadane domenove meno, ukonci funkciu
    
    # prechadza vsetkymy urovnami domeny:
    my $count = 0;
    while ($hostname =~ /\./) { # ak hostname obsahuje bodku
        if ($txt = get_dns_record($hostname, "TXT")) { # ak bol ziskany TXT DNS zaznam hostu
            last;
        }
        ($hostname) = $hostname =~ /\.(.*)$/; # odtrhne z hostname cast po prvu bodku zlava (vratane bodky)
        if (++$count == 2) { last; }  # obmedzenie poctu DNS requestov na 2
    }
    
    # ak bol ziskany TXT DNS zaznam
    if ($txt) {
        # pokusi sa fulltextovo vyhladat plny nazov krajiny:
        my $sel = $dbh->prepare("SELECT code FROM country_codes WHERE ? REGEXP country");
    	$sel->execute($txt); 
    	($country) = get_best_row($sel, $txt);  # ak bola najdena zhoda, ulozi najdenu krajinu do premennej

    	# pokusi sa fulltextovo vyhladat plny nazov mesta:
    	if ($country || $default_country) { 
            my $sel = $dbh->prepare("SELECT city, latitude, longitude FROM geo_city WHERE country LIKE ? AND length(city) > 4 AND ? REGEXP city");
            $sel->execute(($country ? $country : $default_country), $txt);   
            ($city, $lat, $lon) = get_best_row($sel, $txt);  # ak bola najdena zhoda, ulozi najdene mesto a suradnice do premennych
        }
    }
    
    return ($country, $city, $lat, $lon, $txt); 
}


# funkcia lokalizuje IP adresu na zaklade analyzy domenoveho mena
# vstupne parametre: $hostname        - domenove meno uzla
#                    $dbh             - deskriptor databazoveho pripojenia na databazu so suradnicami
#                    $default_country - krajina, ktora uz bola urcena inymi sposobmi
# navratova hodnota: pole (zoznam) lokalizacnych udajov  ($country, $city, $lat, $lon)
#                    $country - krajina lokalizacie
#                    $city    - mesto lokalizacie
#                    $lat     - latitude (zemepisna sirka)
#                    $lon     - longitude (zemepisna dlzka)
sub localize_by_hostname {
    my ($hostname, $dbh, $default_country) = @_;
    my ($country, $city, $lat, $lon, $txt);
    
    if (! $hostname) { return; } # ak nie je zadane domenove meno, ukonci funkciu
    
    my ($postfix) = $hostname =~ /\.(\w+)$/;            # ziska domenu prvej urovne
    my ($hostname_short) = $hostname =~ /(.*)\.\w+$/;   # domenove meno BEZ domeny prvej urovne
    my @words = $hostname_short =~ /([A-Za-z][a-z]+)/g; # ulozi do pola vsetky slova zlozene len z pismen
    my $reg_name = "";
    foreach my $word (@words) {                         # vytvori z tohoto pola vyhladavaci retazec
    	$reg_name .= $word." ";
    }
    
    my @words2 = $hostname_short =~ /([A-Za-z0-9][a-z0-9]+)/g; # ulozi do pola vsetky slova zlozene len z pismen a cislic
    my $reg_name2 = "";
    foreach my $word (@words2) {                               # vytvori z tohoto pola vyhladavaci retazec
    	$reg_name2 .= $word." ";
    }
    
    # pokusi sa najst domenu prvej urovne v tabulke skratiek statov:
	my $sel = $dbh->prepare("SELECT code FROM country_codes WHERE code LIKE ? LIMIT 1");
    $sel->execute($postfix);
    while (my @row = $sel->fetchrow_array) {
    	($country) = @row; # ak existuje zaznam, ulozi skratku statu do premennej $country
    }  
    
    # ak je krajina stale nezistena, pokusi sa fulltextovo vyhladat plny nazov krajiny:
    if (! $country) {
        my $sel = $dbh->prepare("SELECT code FROM country_codes WHERE ? REGEXP country");
    	$sel->execute($reg_name);  
    	($country) = get_best_row($sel, $reg_name);  # ak bola najdena zhoda, ulozi najdenu krajinu do premennej
    }
    
    # ak je krajina stale nezistena, a nie je zistena ani predoslymi sposobmi, pokusi sa fulltextovo vyhladat skratku krajiny:
    if (! $country && ! $default_country) {
        my $sel = $dbh->prepare("SELECT code FROM country_codes WHERE ? REGEXP code");
    	$sel->execute($reg_name2);
    	($country) = get_best_row($sel, $reg_name);  # ak bola najdena zhoda, ulozi najdenu krajinu do premennej
    }
    
    # ak je znama krajina, alebo default krajina, pokusi sa fulltextovo vyhladat plny nazov mesta:
    if ($country || $default_country) { 
        my $sel = $dbh->prepare("SELECT city, latitude, longitude FROM geo_city WHERE country LIKE ? AND length(city) > 4 AND ? REGEXP city ORDER BY length(city) DESC LIMIT 1");
        $sel->execute(($country ? $country : $default_country), $reg_name);
        ($city, $lat, $lon) = get_best_row($sel, $reg_name);  # ak bola najdena zhoda, ulozi najdene mesto a suradnice do premennych
    }
    
    # ak mesto este nebolo najdene a urcena krajina je Slovensko, pokusi sa vyhladat skratku SK mesta
    if (! $city && ($country =~ /sk/i || $default_country =~ /sk/i)) {
        my $sel = $dbh->prepare("SELECT city FROM sk_spz WHERE ? REGEXP spz LIMIT 1");
        $sel->execute($reg_name2);
        ($city) = get_best_row($sel, $reg_name);  # ak bola najdena zhoda, ulozi najdene mesto do premennej
        if ($city) {
            # nacita suradnice urceneho mesta:
            # priamy JOIN na geo_city nie je kvoli efektivite (takto staci 1 match, nie vsetky SK mesta + sk_spz obsahuje duplicitne nazvy miest - Bratislava: BA BL...), tiez aby sa zvysilo skore aj ak suradnice nevieme najst
            my $sel = $dbh->prepare("SELECT latitude, longitude FROM geo_city WHERE country = 'SK' AND city LIKE ? LIMIT 1");
            $sel->execute($city);
            while (my @row = $sel->fetchrow_array) {
                ($lat, $lon) = @row; # ulozi zem. sirku a zem. dlzku z databazy do premennych
    	    }     
        }         
    }
    
    return ($country, $city, $lat, $lon); 
}


# funkcia vrati pozadovanu vetu DNS zaznamu
# vstupne parametre: $host  - IP adresa / domenove meno
#                    $type  - typ pozadovanej vety (A, PTR, TXT, LOC) 
# navratova hodnota: retazec pozadovanej vety DNS zaznamu
sub get_dns_record {
    my ($host, $type) = @_;
    my $res = Net::DNS::Resolver->new(
		tcp_timeout => 1,
		udp_timeout => 1,
	);
    my $query = $res->query($host, $type);
    if ($query) {
        foreach my $rr ($query->answer) {
            if ($rr->type ne $type) { next; } # preskoci nepozadovane vety
            if ($rr->type eq "A") {
                return $rr->address;
            }
            if ($rr->type eq "PTR") {
                return $rr->ptrdname;
            }
            if ($rr->type eq "TXT") {
                return $rr->txtdata;
            }
            if ($rr->type eq "LOC") {
                return $rr->latlon; 
            }
        }
    } 
    return;
}


# vyberie z najdenych databazovych zaznamov ten, ktoreho kluc je v hladanom retazci uvedeny na skorsej pozícii (viacej vlavo) 
# vstupne parametre: $sel				- objekt s ulozenym vysledkom databazoveho dotazu
#                    $search_string     - retazec, v ktorom sa vyhladava zhoda
# navratova hodnota: return_row			- pole hodnot jedneho riadku databazovych udajov
sub get_best_row {
    my ($sel, $search_string) = @_;
    my @return_row;
    
    my $best = 9999; 
    while (my @row = $sel->fetchrow_array) {
    	my $pos = index($search_string, $row[0]);  # vyhlada poziciu mesta v riadku
        if ($pos < $best) {                        # ak je pozicia lepsia ako najlepsia doteraz zistena, aktualizuje udaje
        	@return_row = @row;                    # ulozi mesto, zem. sirku a zem. dlzku z databazy do premennych
            $best = $pos;	
        }
    }    
    
    return @return_row;
}

1;
