#!/usr/bin/perl

package LocIP::Localize;

use strict;
use LocIP::Database;
use LocIP::DNS;
use LocIP::Whois;
use LocIP::Utils;
use Socket;
use DBI;
use threads;
use Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(&localize_host);


# konstanty skore pre jednotlive metody lokalizacie - lokalizacia mesta:
use constant LOCDNS_CITY => 10;
use constant WHOIS_CITY  => 8;
use constant HOSTNAME_CITY => 6;
use constant LOCTXT_CITY => 6;

# konstanty skore pre jednotlive metody lokalizacie - lokalizacia krajiny:
use constant LOCDNS_COUNTRY => 4;
use constant WHOIS_COUNTRY  => 3;
use constant HOSTNAME_COUNTRY => 2;
use constant LOCTXT_COUNTRY => 1;


# funkcia lokalizuje zadanu IP adresu
# vstupne parametre: $host  IP adresa alebo domnov meno, ktor chceme lokalizova
#                    $qeue  ukazovate na rru, do ktorej je vloen vsledok lokalizcie v prpade, e funkcia be vo vlastnom vlkne
#                    $methods  metdy, ktor maj by pouit pri lokalizcii - reazec obsahujci maximlne 4 znaky: l (LOC DNS zznam), w (Whois), d (domnov meno), x (TXT DNS zznam)
# navratove hodnoty: $ip  IP adresa lokalizovanho uzla
#                    $hostname  domnov meno lokalizovanho uzla
#                    $country  krajina lokalizcie
#                    $city  mesto lokalizcie
#                    $lat  zemepisn rka lokalizcie (desatinn slo)
#                    $lon  zemepisn dka lokalizcie (desatinn slo)
#                    $score  dosiahnut skre lokalizcie (cel slo)
#                    $info  textov popis lokalizcie
#                    $success_method  metdy urenia lokalizcie (reazec, formt ako pri vstupnom parametri $methods)
sub localize_host {
    my ($host, $qeue, $methods) = @_;
    my ($ip, $hostname, $country, $city, $lat, $lon, $score, $info, $success_method);
    my ($country_t, $city_t, $lat_t, $lon_t, $info_t);
    
    my $error = 0;
    
    if (length($methods) < 1) { $methods = "lwdx"; }
    
    # ak host je IP adresa:
    if (is_valid_ip($host)) {
        $ip = $host;
        if (is_private_ip($ip)) {
            $error = 1; $score = 0; $info = "Private IP address";
        }
        elsif (is_special_ip($ip)) {
            $error = 1; $score = 0; $info = "Special IP address";
        }
        else {
            # prevedie IP adresu na hostname:
            #$hostname = get_dns_record($ip, "PTR");
            $hostname = gethostbyaddr(inet_aton($host), AF_INET);
        }
    }
    # ak host nie je IP adresa, povazuje sa za hostname
    else {
        $hostname = $host;
        if (! is_valid_hostname($hostname)) { 
            $error = 1; $score = 0; $info = "Bad hostname format"; 
        }
        else { 
            # prevedie hostname na IP:
            my @addresses = gethostbyname($hostname);
            @addresses = map { inet_ntoa($_) } @addresses[4 .. $#addresses];
            $ip = $addresses[0];
            if (! $ip) { $error = 1; $score = 0; $info = "Cannot resolve IP address of host ".$hostname; } 
        }
    }
    
    if (! $error) {
        my $dbh = database_connect();  # vytvori spojenie s SQLite databazou
            
        $score = 0;
        $success_method = "";

        # LOC DNS ZAZNAM:
        if ($methods =~ /l/) {
            ($country, $city, $lat, $lon) = localize_by_loc_dns($hostname, $dbh);
            if ($city) { $score += LOCDNS_CITY; $success_method .= "L "; } elsif ($country) { $score += LOCDNS_COUNTRY; $success_method .= "l "; }
        }
        
        # WHOIS:
        if ($methods =~ /w/ && !($methods =~ /q/ && $city)) {
            ($country_t, $city_t, $lat_t, $lon_t, $info) = localize_by_whois($ip, $dbh);
            ($country, $city, $score, $success_method) = update_loc_details($country, $city, $country_t, $city_t, $score, WHOIS_CITY, WHOIS_COUNTRY, $success_method, "W");
            if ($lat_t && !$lat) { ($lat, $lon) = ($lat_t, $lon_t); }
        }
        
        # DOMENOVE MENO:
        if ($methods =~ /d/ && !($methods =~ /q/ && $city)) {
            ($country_t, $city_t, $lat_t, $lon_t) = localize_by_hostname($hostname, $dbh, $country);
            ($country, $city, $score, $success_method) = update_loc_details($country, $city, $country_t, $city_t, $score, HOSTNAME_CITY, HOSTNAME_COUNTRY, $success_method, "D");
            if ($lat_t && !$lat) { ($lat, $lon) = ($lat_t, $lon_t); }
            if ($lat_t && $score == (HOSTNAME_CITY-(WHOIS_CITY/2))) { ($lat, $lon) = ($lat_t, $lon_t); } # vynimka! ak sa naslo mesto analyzou DNS a dovtedy urcene mesto je urcene len cez Whois
        }
        
        # TXT DNS ZAZNAM:
        if ($methods =~ /x/ && !($methods =~ /q/ && $city)) {
            ($country_t, $city_t, $lat_t, $lon_t, $info_t) = localize_by_txt_dns($hostname, $dbh, $country);
            ($country, $city, $score, $success_method) = update_loc_details($country, $city, $country_t, $city_t, $score, LOCTXT_CITY, LOCTXT_COUNTRY, $success_method, "X");
            if ($lat_t && !$lat) { ($lat, $lon) = ($lat_t, $lon_t); }
            if ($lat_t && $info_t) { $info = $info_t; }
        }
    
    	# ak su nezname suradnice, ale znama krajina - nacita suradnice krajiny:
    	if (!$lat && $country) {
            my $sel = $dbh->prepare("SELECT latitude, longitude FROM geo_city WHERE city = '' AND country LIKE ? LIMIT 1");
        	$sel->execute($country);
        	while (my @row = $sel->fetchrow_array) {
        		($lat, $lon) = @row; # ulozi zem. sirku a zem. dlzku z databazy do premennych
        	}      
        }
        
        # prelozi skratku krajiny na plny nazov:
    	my $sel = $dbh->prepare("SELECT country FROM country_codes WHERE code LIKE ? LIMIT 1");
        $sel->execute($country);
        while (my @row = $sel->fetchrow_array) {
        	($country) = @row; # ulozi plny nazov do premennej $country
        }    
    }
    
    # ak funkcia bezi vo vlakne, vlozi udaje do rury:
    if (threads->self()->tid() > 0) {
        lock($qeue);
        $qeue->enqueue($ip, $hostname, $country, $city, $lat, $lon, $score, $info, $success_method);
    }
    return ($ip, $hostname, $country, $city, $lat, $lon, $score, $info, $success_method);
}


# privatna funkcia pre spracovanie vysledkov lokalizacie danou metodou
sub update_loc_details {
    my ($country, $city, $country_t, $city_t, $score, $city_score, $country_score, $success_method, $method_name) = @_;
    # print $country_t, " ", $city_t, "", $country_score, "\n"; #DEBUG
    # ak bolo zistene mesto lokalizacie:
    if ($city_t) {
        # ak sa mesto zhoduje s uz urcenym, zvysi sa skore:
        if ($city_t eq $city) { 
            $score += $city_score; 
        }
        # ak sa mesto nezhoduje s uz urcenym, znizi sa skore:
        elsif ($city) { 
            # vynimka! ak sa naslo mesto analyzou hostname a dovtedy urcene mesto je urcene len cez Whois
            if ($score == WHOIS_CITY && $method_name == "D") {
                $city = $city_t;  # pouzije sa mesto a krajina zistena z domenoveho mena
                if ($country_t) { $country = $country_t; }
                $score = $city_score - (WHOIS_CITY/2);      # score = score hostname - 1/2 score Whois 
            }
            else {
                $score -= int($city_score/2); 
            }
        }
        # ak inou metodou nebolo urcene mesto, ulozi mesto a krajinu a zvysi skore:
        else {
            ($country, $city) = ($country_t, $city_t);
            $score += $city_score;
        }
        
        $success_method .= $method_name." ";
    }
    # ak nebolo zistene mesto, ale bola zistena krajina:
    elsif ($country_t) {
        # ak sa krajina zhoduje s uz urcenou, zvysi sa skore:
        if ($country_t eq $country) { 
            $score += $country_score; 
        }
        # ak sa krajina nezhoduje s uz urcenou, znizi sa skore:
        elsif ($country) { 
            $score -= int($country_score/2);
        }
        # ak inou metodou nebola urcena krajina, ulozi krajinu a zvysi skore:
        else {
            $country = $country_t;
            $score += $country_score;
        }
        
        $success_method .= lc($method_name)." ";
    }
    return ($country, $city, $score, $success_method);
}

1;
