#!/usr/bin/perl

package LocIP::Utils;

use strict;
use Math::Trig;
use Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(&get_distance &is_valid_ip &is_valid_hostname &is_private_ip &is_special_ip &trim &round &latlon2readable &get_precision);


# funkcia vrati vzdialenost medzi dvoma bodmi zadanymi v suradniciach
# Haversineho formula: http://www.movable-type.co.uk/scripts/latlong.html
# vstupne parametre: $lat1 - zemepisna sirka (latitude) prveho bodu
#                    $lon1 - zemepisna dlzka (longitude) prveho bodu
#                    $lat2 - zemepisna sirka (latitude) druheho bodu
#                    $lon2 - zemepisna dlzka (longitude) druheho bodu
# navratova hodnota: vzdialenost v km
sub get_distance {
    my ($lat1, $lon1, $lat2, $lon2) = @_;
    
    my $R = 6371; # polomer Zeme v km
    my $dLat = deg2rad($lat2-$lat1);
    my $dLon = deg2rad($lon2-$lon1); 
    my $a = sin($dLat/2) * sin($dLat/2) + cos(deg2rad($lat1)) * cos(deg2rad($lat2)) *  sin($dLon/2) * sin($dLon/2); 
    my $c = 2 * atan2(sqrt($a), sqrt(1-$a)); 
    my $d = $R * $c;
    
    return $d;
}


# funkcia overi platnost formatu IP adresy
# vstupne parametre: $ip - IP adresa
# navratova hodnota: 1 ak je IP adresa v platnom formate, 0 ak nie je
sub is_valid_ip {
    my ($ip) = @_;
    return $ip =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/;
}


# funkcia overi platnost formatu domenoveho mena
# vstupne parametre: $hostname - domenove meno
# navratova hodnota: 1 ak je domenove meno v planom formate, 0 ak nie je
sub is_valid_hostname {
    my ($hostname) = @_;
    return $hostname =~ /^(\w|\.|-)+\.(\w|\.|-)+$/;
}


# funkcia ozisti, ci je zadana IP adresa privatna
# vstupne parametre: $ip - IP adresa
# navratova hodnota: 1 ak je IP adresa privatna, 0 ak nie je
sub is_private_ip {
    my ($ip) = @_;
    if ($ip =~ /^10\./) { return 1; }                               # 10.0.0.0 - 10.255.255.255 
    if ($ip =~ /^172\.(1[6-9]|2\d|3[01])\./) { return 1; }          # 172.16.0.0 - 172.31.255.255 
    if ($ip =~ /^192\.168\./) { return 1; }                         # 192.168.0.0 - 192.168.255.255 
    return 0;
}


# funkcia ozisti, ci je zadana IP adresa specialna (loopback, multicast)
# vstupne parametre: $ip - IP adresa
# navratova hodnota: 1 ak je IP adresa specialna, 0 ak nie je
sub is_special_ip {
    my ($ip) = @_;
    if ($ip =~ /^127\./) { return 1; }                              # 127.0.0.0 - 127.255.255.255 
    if ($ip =~ /^2(2[4-9]|3\d)\./) { return 1; }                  # 224.0.0.0 - 239.255.255.255
    if ($ip =~ /^2[45]\d\./) { return 1; }                          # 240.0.0.0 - 255.255.255.255 
    return 0;
}


# funkcia odstrani biele znaky zo zaciatku a konca retazca
# vstupne parametre: $string - vstupny retazec
# navratova hodnota: retazec s odstranenymi bielymi znakmi na zaciatku a konci retazca
sub trim($) {
	my $string = shift;
	$string =~ s/^\s+//;
	$string =~ s/\s+$//;
	return $string;
}


# funkcia zaokruhli desatinne cislo na zadany pocet desatinnych miest
# vstupne parametre: $number - desatinne cislo
#                    $dec    - pocet desatinnych miest
# navratova hodnota: zaokruhlene cislo
sub round {
    my $number = shift || 0;
    my $dec = 10 ** (shift || 0);
    return int( $dec * $number + .5 * ($number <=> 0)) / $dec;
}


# funkcia premeni suradnice (uhol v stupnoch - desatinne cislo) na stupne, minuty, sekundy a smer
# vstupne parametre: $decimal - uhol v stupnoch - desatinne cislo
# navratova hodnota: pole (zoznam):
#                    $degrees - stupne
#                    $minutes - minuty
#                    $seconds - sekundy
#                    $sign    - znamienko (kladne 1, zaporne -1)
sub decimal2dms {
    my ($decimal) = @_;

    my $sign = $decimal <=> 0;
    my $degrees = int($decimal);

    # convert decimal part to minutes
    my $dec_min = abs($decimal - $degrees) * 60;
    my $minutes = int($dec_min);
    my $seconds = ($dec_min - $minutes) * 60;

    return (abs($degrees), $minutes, $seconds, $sign);
}


# funkcia premeni suradnice (uhol v stupnoch - desatinne cislo) na retazec formatu stupne, minuty, sekundy polohula (severna/juzna, vychodna/zapadna)
# vstupne parametre: $lat  - zemepisna sirka (uhol v stupnoch - desatinne cislo)
#                    $lon  - zemepisna dlzka (uhol v stupnoch - desatinne cislo)
#                    $html - pouzit specialne html znacky (&deg; &#39; &quot;)
# navratova hodnota: pole (zoznam):
#                    $lat_r - zemepisna sirka formatu stupne, minuty, sekundy polohula (severna/juzna, vychodna/zapadna)
#                    $lon_r - zemepisna dlzka formatu stupne, minuty, sekundy polohula (severna/juzna, vychodna/zapadna)
sub latlon2readable {
    my ($lat, $lon, $html) = @_;
    my ($degrees, $minutes, $seconds, $sign, $lat_r, $lon_r);

    if ($lat == 0 && $lon == 0) { return; }   
    
    my $deg_sign = ($^O =~ /mswin/i) ? 248 : 176 ;
 
    ($degrees, $minutes, $seconds, $sign) = decimal2dms($lat);
    if ($html) {
        $lat_r = sprintf("%d&deg; %02d&#39; %02d&quot; %s", $degrees, $minutes, $seconds, ($sign > 0 ? "N" : "S"));
    }
    else {
        $lat_r = sprintf("%d%c %02d' %02d\" %s", $degrees, $deg_sign, $minutes, $seconds, ($sign > 0 ? "N" : "S"));
    }
    
    ($degrees, $minutes, $seconds, $sign) = decimal2dms($lon);
    if ($html) {
        $lon_r = sprintf("%d&deg; %02d&#39; %02d&quot; %s", $degrees, $minutes, $seconds, ($sign > 0 ? "E" : "W"));
    }
    else {
        $lon_r = sprintf("%d%c %02d' %02d\" %s", $degrees, $deg_sign, $minutes, $seconds, ($sign > 0 ? "E" : "W"));
    }
    
    return ($lat_r, $lon_r);  
}


# funkcia premeni skore na slovnu popis presnosti lokalizacie
# vstupne parametre: $score - vyska skore lokalizacie
# navratova hodnota: retazec obsahujuci popis presnosti lokalizacie
sub get_precision {
    my ($score) = @_;
    if ($score < 6)     { return "VL"; }
    elsif ($score < 14) { return "L"; }
    elsif ($score < 18) { return "H"; }
    else                { return "VH"; }
}
