#!/usr/bin/perl

BEGIN {
    # prida do @INC adresar, v ktorom sa nachadza script:
    require lib;
    my ($dir) = $0 =~ /(.*)\//;
    if ($dir) { lib->import($dir); }
}

use strict;
use LocIP::Localize;
use LocIP::Utils;
use Getopt::Std;
use threads;
use Thread::Queue;
use Sys::HostIP;
use utf8;

use constant MAX_THREADS => 10; # maximalny pocet vlakien, v ktorych moze bezat lokalizacia


my (%opts, @hosts);
getopts("hmftvqlwdxaspb",\%opts);

my $methods = ""; # metody, ktorymi sa bude vyhladavat
if ($opts{"l"}) { $methods .= "l"; }
if ($opts{"w"}) { $methods .= "w"; }
if ($opts{"d"}) { $methods .= "d"; }
if ($opts{"x"}) { $methods .= "x"; }
if (! $methods) { $methods = "lwdx"; } # default: vsetky metody
if ($opts{"q"}) { $methods .= "q"; } # rychle vyhladanie

if ($opts{"h"}) {
    print_help();
}
elsif ($opts{"m"}) {
    push @hosts, Sys::HostIP->ip; 
}
elsif ($opts{"f"}) {
	foreach my $filename (@ARGV) {
		if (! open(INPUTFILE, $filename)) {
			print STDERR "Error by opening file '$filename'\n";
			next;
		}
        while (my $line = <INPUTFILE>) {
            chomp($line);
            my @newhosts = ($line =~ /(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/g); # ulozi do newhosts vsetky IP na riadku
            if (@newhosts > 0) { 
                push @hosts, @newhosts;
            }
        }
        close(INPUTFILE);
	}
	if (! @ARGV) {
        while (my $line = <>) {
            chomp($line);
            push @hosts, split(/\s/, $line);
        }
    }
}
elsif ($opts{"t"}) {
    if (! @ARGV) {
        print_help();
        exit;
    }
    
    if ($^O =~ /mswin/i) {
        open TR, "tracert $ARGV[0] |";    # OS Windows
    }
    else {
        open TR, "traceroute -q 5 $ARGV[0] |"; # OS nie Windows
    }
    
    @hosts = ();
    # vsetky IP adresy z vystupu traceroute vlozi do pola hosts:
    while (my $line = <TR>) {
        if ($line !~ /^\s*\d/) { next; } # len riadky zacinajuce cislicou (hop-y)
        my ($host, $host2, $host3, $host4, $host5) = ($line =~ /(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/g); # vzdy prva IP adresa z riadku
        if ($host) {
            if (grep(/$host/, @hosts) && $host2) { $host = $host2; }
            if (grep(/$host/, @hosts) && $host3) { $host = $host3; }
            if (grep(/$host/, @hosts) && $host4) { $host = $host4; }
            if (grep(/$host/, @hosts) && $host5) { $host = $host5; }
            push @hosts, $host;
            #print $host, " ", $line;
        }
    }
    close TR;
}
else {
    if (! @ARGV) { print_help(); }
	push @hosts, @ARGV;
}


my @vars = ("ip", "hostname", "country", "city", "lat", "lon", "score", "info", "method"); # pozadovane lokalizacne udaje:
my %hostinfos; # asociativne pole asociativnych poli obsahujucich informacie o IP adresach (kluc - IP adresa)

# spracovanie lokalizacie viacerych zadanych hostov vo vlaknach:
if (@hosts > 1) {
    my $queue = Thread::Queue->new();
    my @hosts_c = @hosts;
    my $threads_count = 0;
    
    while (@hosts_c > 0 || $threads_count > 0) { # pokial este existuju nelokalizovane hosty alebo bezi nejake vlakno
        # ak este existuju hosty cakajuce na lokalizaciu a nebol prekroceny max. pocet vlakien:
        if ($threads_count < MAX_THREADS && @hosts_c > 0) {
            my $host = shift(@hosts_c);
            my $thr = threads->create(\&localize_host, $host, $queue, $methods); # vytvori nove vlakno v ktorom sa pokusi lokalizovat host
            #$thr->detach(); # necaka sa na ukoncenie vlakna
            $threads_count++;
        }
        # ak bol prekroceny max. pocet vlakien, alebo uz nie je dalsi host cakajuci na lokalizaciu:
        else {
            my %hostinf;
            # pocka na ukoncenie nejakeho vlakna a nacita lokalizacne udaje z rury:
            foreach my $varname (@vars) {
                $hostinf{$varname} = $queue->dequeue();
            }
            $threads_count--;
            #print_localization(%hostinf); # vypise info o lokalizacii
            # ulozi lokalizaciu do asociativneho pola:
            if (grep(/$hostinf{"ip"}/, @hosts)) { # bolo zadane hostname alebo IP adresa?
                $hostinfos{$hostinf{"ip"}} = \%hostinf; # kluc je IP adresa
            }
            else {
                $hostinfos{$hostinf{"hostname"}} = \%hostinf; # kluc je hostname
            }
        }
    }
    
    # Join kazdeho beziaceho vlakna
    foreach my $thr (threads->list) { 
        # join kazdeho vlakna okram hlavneho a vlastneho
        if ($thr->tid && !threads::equal($thr, threads->self)) { 
            $thr->join; 
        } 
    }
}
# spracovanie lokalizacie jedneho zadaneho hostu:
elsif (@hosts == 1) {
    my %hostinf;
    ($hostinf{"ip"}, $hostinf{"hostname"}, $hostinf{"country"}, $hostinf{"city"}, $hostinf{"lat"}, $hostinf{"lon"}, $hostinf{"score"}, $hostinf{"info"}, $hostinf{"method"}) = localize_host($hosts[0], "", $methods); # pokusi sa lokalizovat host
	#print_localization(%hostinf); # vypise info o lokalizacii
	$hostinfos{$hosts[0]} = \%hostinf; # ulozi lokalizaciu do asociativneho pola
}

# vypise informacie o IP adresach v poradi ako boli zadane:
if (@hosts) {
    #vzdialenost medzi prvymi dvoma zadanymi IP adresami:
    if ($opts{"b"} && @hosts > 1) {
        my %hostinf1 = %{$hostinfos{$hosts[0]}};
        my %hostinf2 = %{$hostinfos{$hosts[1]}};
        print round(get_distance($hostinf1{"lat"}, $hostinf1{"lon"}, $hostinf2{"lat"}, $hostinf2{"lon"}), 1), "\n";
        exit;
    }

    print_header();
    my %last;
    foreach my $host (@hosts) {
        my %hostinf = %{$hostinfos{$host}}; # nacita lokalizacne udaje o danej IP adrese
        
        # vypocet vzdialenosti od predchadzajucej IP adresy:
        if ($opts{"v"}) {
            if (%last) {
                $hostinf{"distance"} = round(get_distance($last{"lat"}, $last{"lon"}, $hostinf{"lat"}, $hostinf{"lon"}), 1);
            }
            %last = %hostinf;
        }
        print_localization(%hostinf); # vypise info o lokalizacii
    }
}




sub print_help {
    print "LocIP - IP addresses localization tool\t(C) 2009 Rastislav Szabo\n\n";
	print "Usage:\tlocip.pl [-options] [-f [filename1 filename2 ...]] [host1 host2 ...]\n";
	print "Options:\n";
	print "\t-h                           : application help\n";
	print "\t-m                           : localize IP address of this computer\n";
	print "\t-t                           : localize hops on route to entered host (traceroute)\n";
	print "\t-f [filename1 filename2 ...] : localize IP addresses from file(s), STDIN if no filename specified\n";
    print "\thost1 host2 ...              : IP addresses/hostnames to localize\n";
    print "\n";
    print "\t-v                           : calculate distance between each IP and previous entered IP (in kilometers)\n";
    print "\t-b                           : calculate distance between first two entered IP addresses (in kilometers)\n";
    print "\t-q                           : quick localization\n";
    print "\t-l                           : localize only using LOC DNS entry\n";
    print "\t-w                           : localize only using Whois\n";
    print "\t-d                           : localize only using hostname\n";
    print "\t-x                           : localize only using TXT DNS entry\n";
    print "\n";
    print "\t-a                           : print all information about IP addresses\n";
    print "\t-s                           : print short information about IP addresses\n";
    print "\t-p                           : print coordinates in floating point number format (negative mens southern latitude / western longitude)\n";
    print "\n";
    print "Coordinates format: N - north latitude, S - south latitude, W - west longitude, E - east longitude \n";
    print "Score format:  VL - very low, L - low, H - high, VH - very high\n";
    print "Method format: L - LOC DNS entry, W - Whois, D - domain (hostname), X - TXT DNS entry. Lowercase if only country found.\n";
}


sub print_header {
    if ($opts{"a"}) {
        if ($opts{"v"}) {
            printf "\n%-15s %-40s %-15s %-20s %-14s %-14s %-8s %-5s %-7s %-10s\n", "IP address", "Hostname", "Country", "City", "Latitude", "Longitude", "Distance", "Score", "Method", "Info";
        }
        else {
            printf "\n%-15s %-40s %-15s %-20s %-14s %-14s %-5s %-7s %-10s\n", "IP address", "Hostname", "Country", "City", "Latitude", "Longitude", "Score", "Method", "Info";
        }
        print "=" x 150, "\n";
    }
    elsif ($opts{"s"}) {
        if ($opts{"v"}) {
            printf "\n%-15s %-15s %-20s %-8s %-10s\n", "IP address", "Country", "City", "Distance", "Info";
        }
        else {
            printf "\n%-15s %-15s %-20s %-10s\n", "IP address", "Country", "City", "Info";
        }
        print "=" x 80, "\n";
    }
    else {
        if ($opts{"v"}) {
            printf "\n%-15s %-15s %-20s %-14s %-14s %-8s %-5s %-7s %-10s\n", "IP address", "Country", "City", "Latitude", "Longitude", "Distance", "Score", "Method", "Info";
        }
        else {
            printf "\n%-15s %-15s %-20s %-14s %-14s %-5s %-7s %-10s\n", "IP address", "Country", "City", "Latitude", "Longitude", "Score", "Method", "Info";
        }
        print "=" x 125, "\n";
    }
}


sub print_localization {
	my (%hostinf) = @_;
	
	# premeni suradnice do formatu: stupne, minuty, sekundy:
	if (! $opts{"p"}) {
	   ($hostinf{"lat"}, $hostinf{"lon"}) = latlon2readable($hostinf{"lat"}, $hostinf{"lon"});
	}
	
	if ($opts{"a"}) {
        if ($opts{"v"}) {
            printf "%-15s %-40s %-15s %-20s %-14.14s %-14.14s %-8.8s %-2s %-2s %-7s %-10s\n", $hostinf{"ip"}, $hostinf{"hostname"}, $hostinf{"country"}, $hostinf{"city"}, $hostinf{"lat"}, $hostinf{"lon"}, $hostinf{"distance"}, $hostinf{"score"}, get_precision($hostinf{"score"}), $hostinf{"method"}, $hostinf{"info"};
        }
        else {
            printf "%-15s %-40s %-15s %-20s %-14.14s %-14.14s %-2s %-2s %-7s %-10s\n", $hostinf{"ip"}, $hostinf{"hostname"}, $hostinf{"country"}, $hostinf{"city"}, $hostinf{"lat"}, $hostinf{"lon"}, $hostinf{"score"}, get_precision($hostinf{"score"}), $hostinf{"method"}, $hostinf{"info"};
        }
    }
    elsif ($opts{"s"}) {
        if ($opts{"v"}) {
            printf "%-15s %-15s %-20s %-8.8s %-10s\n", $hostinf{"ip"}, $hostinf{"country"}, $hostinf{"city"}, $hostinf{"distance"}, $hostinf{"info"}; 
        }
        else {
            printf "%-15s %-15s %-20s %-10s\n", $hostinf{"ip"}, $hostinf{"country"}, $hostinf{"city"}, $hostinf{"info"};  
        } 
    }
    else {
        if ($opts{"v"}) {
            printf "%-15s %-15s %-20s %-14.14s %-14.14s %-8.8s %-2s %-2s %-7s %-10s\n", $hostinf{"ip"}, $hostinf{"country"}, $hostinf{"city"}, $hostinf{"lat"}, $hostinf{"lon"}, $hostinf{"distance"}, $hostinf{"score"}, get_precision($hostinf{"score"}), $hostinf{"method"}, $hostinf{"info"};
        }
        else{
            printf "%-15s %-15s %-20s %-14.14s %-14.14s %-2s %-2s %-7s %-10s\n", $hostinf{"ip"}, $hostinf{"country"}, $hostinf{"city"}, $hostinf{"lat"}, $hostinf{"lon"}, $hostinf{"score"}, get_precision($hostinf{"score"}), $hostinf{"method"}, $hostinf{"info"};
        }
    }
}
